require "log"
require "math"
require "globals"
require "WorldObject"
require "Bullet"
require "Missile"
require "Explosion"


-- fighter class
Fighter = inheritsFrom(WorldObject) 


function Fighter:new(world, x,y, name)
	local self = self:create()
	
	self.image_map = nil
	self.fighter_quad = nil
	self.burn_quad = nil
	self.burn_offset_y = 0
	self.pilot = nil

	self.burn_animation_timer = 0
	self.thrust = 5
	self.turnThrust = 0.05
	self.mass = 10
	self.boundingRadius = 37
	self.w = 74
	self.h = 47
	self.xVel = 0
	self.yVel = 0
	self.rVel = 0
	self.heading = 0
	self.collided = false
	self.centerHealth = 100
	self.portHealth = 100
	self.starboardHealth = 100
	self.laserTimer = 0
	self.laserSwitch = 1
	self.targetObject = nil
	self.energy = 100
	self.missilesLeft = 10
	self.missileTimer = 0
	self.missileCycleTimer = 0

	self.detectedObjects = {}
	self.targetedObject = nil
	self.closestDetectedObject = nil

	log(" Fighter spawning: " .. name)
	self.world = world
	self.x = x or 0
	self.y = y or 0
	self.name = name or "Anonymous Fighter"
	self.engineStates = { 0,0,0,0 } --FL,FR,RR,RL (Clockwise 1-4). 0 is off, 10 is full
	self.engineGlow = { 0, 0, 0, 0 }
	self:loadSprite() --Loads the sprite into the above vars
	self:defineHitCircles()
	self.status = 1 ---1: exploding
				--0: Destroyed
				--1: Good
	log(" Fighter loaded")
	return self
end

--Return hit circles for collider
--A circle is:
--x, y, r 
--The X and Y are translated appropriately depending on the object's heading/position
--(rotate around the center first, then add in the object's XY)
function Fighter:getCircles() --for collider
	--Rotation around the center of the object:
	--x' = x cos f - y sin f
	--y' = y cos f + x sin f
	local circle_list = {}
	--go thru hit circles, rotate them around the center, add the object's X,Y, package them
	--in a new table, return it
	for i,circle in ipairs(self.hitCircles) do
		local temp_circle = {}
		temp_circle.x = self.x+((circle.x*math.cos(self.heading)) + (circle.y*math.sin(self.heading)))
		temp_circle.y = self.y+((circle.y*math.cos(self.heading)) - (circle.x*math.sin(self.heading)))
		temp_circle.r = circle.r
		circle_list[#circle_list+1] = temp_circle --Add to temp circle listing
	end
	return circle_list
end

--Set up the fighter's hit circles (relative center of object)
function Fighter:defineHitCircles()
	local circle_list = {}
	local fuselage_circle = {}
		fuselage_circle.x = 0
		fuselage_circle.y = 8
		fuselage_circle.r = 13
	local port_nacel_circle = {}
		port_nacel_circle.x = -19
		port_nacel_circle.y = 0
		port_nacel_circle.r = 11
	local starboard_nacel_circle = {}
		starboard_nacel_circle.x = 19
		starboard_nacel_circle.y = 0
		starboard_nacel_circle.r = 11
	circle_list[1] = fuselage_circle
	circle_list[2] = port_nacel_circle
	circle_list[3] = starboard_nacel_circle
	self.hitCircles = circle_list
end
--Activate engines 1-4
function Fighter:burnEngine(engine, dt)
	--Check to see if the engine functions, etc.
	if self.energy <=0 then	return end
	if not dt then log("Called Fighter:burnengine() without dt", 10); return end  
	self.energy = self.energy - 10*dt	
	self.engineStates[engine] = 10
	if self.engineGlow[engine] < 12 then self.engineGlow[engine] = self.engineGlow[engine] + 14*dt end

end
function Fighter:turnEnginesOff(dt)
	-- Reset all the engines states. They'll be turned on again by the input check
	for i=1, 4 do
		self.engineStates[i] = 0
		if self.engineGlow[i] > 0 then self.engineGlow[i] = self.engineGlow[i] - 8*dt end 
	end
end

function Fighter.preLoad()
	global.fighterImage = love.graphics.newImage(global.fighterImageName)
end

function Fighter:loadSprite()
	self.image_map = global.fighterImage
    local image_map_w, image_map_h = self.image_map:getWidth(), self.image_map:getHeight()

    local fighter_sprite_w, fighter_sprite_h = self.w, self.h
	self.fighter_quad = love.graphics.newQuad(0, 0, fighter_sprite_w, fighter_sprite_h,
											image_map_w, image_map_h)

	local burn_sprite_w, burn_sprite_h = 13,13
	local burn_sprite_left, burn_sprite_top = 77, 0
	self.burn_quad = love.graphics.newQuad(burn_sprite_left, burn_sprite_top,
											burn_sprite_w, burn_sprite_h,
											image_map_w, image_map_h)
--	local burn_sprite_w, burn_sprite_h = 26,29
--	local burn_sprite_left, burn_sprite_top = 74, 13
--	self.burn_quad = love.graphics.newQuad(burn_sprite_left, burn_sprite_top,
--											burn_sprite_w, burn_sprite_h,
--											image_map_w, image_map_h)

	if self.image_map == nil then log("*Error loading fighter image.")
		else log(" Done loading fighter image.") end
end

function Fighter:draw()

	--Camera has translated us to the middle of the object
	--camera has rotated to player's heading
	--Rotate around the center
	love.graphics.rotate(self.heading)

	--Move to top left of the object for drawing
	love.graphics.translate(-self.w/2,-self.h/2) 

	--Draw the fighter
	love.graphics.drawq(self.image_map, self.fighter_quad, 0, 0, 0)
	love.graphics.setColorMode("modulate")
	love.graphics.setBlendMode("additive")	
	if self.engineGlow[1] > 0 then --Front left
		love.graphics.setColor(255,255,255,self.engineGlow[1]*16)
		love.graphics.drawq(self.image_map, self.burn_quad, 5, 3)
	end
	if self.engineGlow[2] > 0 then --Front Right
		love.graphics.setColor(255,255,255,self.engineGlow[2]*16)
		love.graphics.drawq(self.image_map, self.burn_quad, 55, 3)
	end
	if self.engineGlow[3] > 0 then --Rear Right
		love.graphics.setColor(255,255,255,self.engineGlow[3]*16)
		love.graphics.drawq(self.image_map, self.burn_quad, 55, 35)
	end
	if self.engineGlow[4] > 0 then --Rear Left
	love.graphics.setColor(255,255,255,self.engineGlow[4]*16)
		love.graphics.drawq(self.image_map, self.burn_quad, 5, 35)
	end
	love.graphics.setColorMode("replace")
end

function Fighter:setPilot(ai)
	self.pilot = ai
end

function Fighter:update(dt)
	self:turnEnginesOff(dt)
	self.burn_animation_timer = self.burn_animation_timer + dt
	if self.burn_animation_timer > 0.002 then 

		self.burn_offset_y = (self.burn_offset_y + 13) % (3*13)
		self.burn_quad:setViewport(77, self.burn_offset_y, 13, 13)
		self.burn_animation_timer = 0
	end
	if self.energy < 100 then self.energy = self.energy + 10*dt end
	if self.pilot ~= nil then
		 self.pilot:update(dt) end
	self:updateMovement(dt)

end

function Fighter:updateMovement(dt)
	local thrust = 0
	local r_thrust = 0
	--Forward
	thrust = self.thrust*(self.engineStates[3] + self.engineStates[4])
	--Back
	thrust = thrust- (self.thrust*(self.engineStates[1] + self.engineStates[2]))
	--Right
	r_thrust = self.turnThrust*(self.engineStates[2] + self.engineStates[4])
	--Left
	r_thrust = r_thrust- (self.turnThrust*(self.engineStates[1] + self.engineStates[3]))

	--Accelerations
	self.xVel = self.xVel + (math.sin(self.heading)*(thrust*dt))
	self.yVel = self.yVel + (math.cos(self.heading)*(thrust*dt)) --Remember Y is always inverted

	self.rVel = self.rVel + (r_thrust*dt)

	--Displacements
	self.heading = self.heading + (self.rVel*dt)
	-- don't let the heading get bigger than one full circle
	if self.heading > 2*math.pi then self.heading = 0 + (self.heading%(2*math.pi)) end
	-- no fucking negative headings
	if self.heading < 0 then self.heading = (2*math.pi)+self.heading end
	self.x = self.x + self.xVel*dt
	self.y = self.y + self.yVel*dt

end

function Fighter:receiveCollision(hit_list, world_object)
	for i,hit_circle in ipairs(hit_list) do
		--1: center 2: port 3: starboard
		if hit_circle == 1 then self:centerHit(world_object)
		elseif hit_circle == 2 then self:centerHit(world_object)
		elseif hit_circle == 3 then self:centerHit(world_object) end
	end

end

function Fighter:centerHit(world_object) 
	-- Kinetic damage: absolute relative velocity multiplied by mass, multiplied by some global
	-- factor
	local dv = math.abs((world_object.xVel+world_object.yVel)-(self.xVel+self.yVel))
	local damage = world_object:extraDamage(self)
	print (self.name .. ": Damage: " .. damage .. " from " .. world_object.name)
	damage = damage + (dv * world_object.mass) * global.kineticFactor
--	print ("Damage: " .. damage .. '(' .. dv .. "*" .. world_object.mass .. "*" .. global.kineticFactor .. ")")
	self.centerHealth = self.centerHealth - damage
	if self.centerHealth < 0 then self:explode() end
end

function Fighter:explode()
	self.status = -1
end

function Fighter:outOfBounds()
	--if there's a pilot and he has a name and that name is auto return then dont bother
	if self.pilot and self.pilot.name and self.pilot.name == "Auto Return" then return end
	print ("FIGHTER OUT OF BOUNDS. AUTOPILOT ENGAGED")
	local old_pilot = self.pilot
	self.pilot = AIBasic:create()
	self.pilot:new(self.world, self, "Auto Return")
	self.pilot:q({"returnToAO"})
	self.pilot:q({"giveControlTo", old_pilot} )
end

function Fighter:ai()
	--if there's a pilot and he has a name and that name is auto return then dont bother
	if self.pilot and self.pilot.name and self.pilot.name == "Auto Return" then return end
	print ("M. AUTOPILOT ENGAGED")
	print ("rVel: " .. self.rVel)
	local old_pilot = self.pilot
	self.pilot = AIBasic:create()
	self.pilot:new(self.world, self, "Auto Return")
	self.pilot:q({"turnToHeading", 100, 0})
	self.pilot:q({"giveControlTo", old_pilot} )
end

function Fighter:shoot()
	if love.timer.getTime() - self.laserTimer < 0.1 or self.energy < 10 then return end
	self.energy = self.energy -10
	self.laserTimer = love.timer.getTime()
	--create the bullet out in front
	self.laserSwitch = self.laserSwitch * -1
	local x_create = self.x + (math.cos(self.heading)*(self.laserSwitch*10)) + (40*math.sin(self.heading))
	local y_create = self.y + (math.cos(self.heading)*40)- ((self.laserSwitch*10)*math.sin(self.heading))

	--temp_circle.x = self.x+((circle.x*math.cos(self.heading)) + (circle.y*math.sin(self.heading)))
	--temp_circle.y = self.y+((circle.y*math.cos(self.heading)) - (circle.x*math.sin(self.heading)))

	self.world:spawnBullet(self, x_create, y_create, self.xVel, self.yVel, self.heading)

end

function Fighter:readyMissile()
	self.missileTimer = love.timer.getTime()
end

function Fighter:launchMissile()
	if love.timer.getTime() - self.missileCycleTimer < 0.1 then return end
	if self.missilesLeft < 1 then return end
	self.missileCycleTimer = love.timer.getTime()
	local x_create = self.x + (math.cos(self.heading)) + (40*math.sin(self.heading))
	local y_create = self.y + (math.cos(self.heading)*40)- (math.sin(self.heading))
	self.world:spawnMissile(self, x_create, y_create, self.xVel, self.yVel, self.heading, self.rVel, (love.timer.getTime() - self.missileTimer) * 2)	
	self.missilesLeft = self.missilesLeft -1
end

function Fighter:getEmission()
	return 100-self.energy+20
end


--
--Targeting computer
--
function Fighter:targetClosestThreat()
	if self.closestDetectedObject then self.targetedObject = self.closestDetectedObject end
end

function Fighter:scan()
	self.detectedObjects = {} --clean out the table
	self.closestDetectedObject = nil
	local distance = 0
	local emission = 0
	local array_size = 0
	local reading = nil
	for i,world_object in pairs(self.world.energyObjects) do
		if world_object ~= self then
			distance = distanceToXY(self.x, self.y, world_object.x, world_object.y)
			emission = world_object:getEmission() / (distance / global.energyFade)			
			reading = { world_object, emission, distance }
			self.detectedObjects[i] = reading

			if distance < global.screenRadius then --if more or less on screen...
				
				if not self.closestDetectedObject or distance < self.detectedObjects[self.closestDetectedObject][3] then --and closest object 
					self.closestDetectedObject = i
				end
			end 
		end 	
	end
end


