
-- World class
require "WorldObject"
require "SmokeTrail.lua"
require "ExplosionB.lua"
require "Bullet"
require "Missile"
require "Fighter"
require "Asteroid"
require "Marker"
require "collisions"
require "level"

World = {
	-- Class variables
	fighter1 = 0,
	radius = 3000,
	warnRadius = 0,
	-- An object list for each layer
	worldObjects = {},
	--explosionObjects = {},
	energyObjects = {},
	spawnPoints = {},
	background = nil,
	backgroundQuad = nil
}
World.__index = World 

--Constructor
function World:new(radius)
	local o = {}		--The instanciation of the object
	setmetatable(o, self)
    -- initialize collider library
	
	self:loadBackground()
	self.radius = radius or 3000
	self.warnRadius = self.radius * .80
	loadLevel(self, "levels/level1.lvl")
	return o
end



function World:loadBackground()
	self.background = global.backgroundImage
	self.backgroundQuad = love.graphics.newQuad( 0, 0, 800, 800, self.background:getWidth(), self.background:getHeight() )
	if self.background == nil then log("*Error loading background image.")
		else log(" Done loading background image.") end
end

function World.preLoad()
	global.backgroundImage = love.graphics.newImage(global.backgroundImageName)

end

function World:getWorldObjects()
	return self.worldObjects
end

function World:getDrawLists()
	local draw_list = {}
	draw_list[1] = self.worldObjects
	return draw_list
end


function World:update(dt)	
	World:updateWorldObjects(dt)

end

function World:spawnExplosion(x, y, xVel, yVel, explosion)
	local object = explosion:new(x,y,xVel,yVel)
	self.worldObjects[#self.worldObjects+1] = object
end

function World:updateWorldObjects(dt)
	for i,world_object in ipairs(self.worldObjects) do
		--Check for exploding objects
		if world_object.status == -1 then --ExplosionA
			self:spawnExplosion(world_object.x,world_object.y,world_object.xVel,world_object.yVel, Explosion)
			table.remove(self.worldObjects, i) --skip all other updates for this object
			if world_object.energyIndex > -1 then self.energyObjects[world_object.energyIndex] = nil end -- it may not be in there, but whatever
		elseif world_object.status == -2 then --ExplosionB
			self:spawnExplosion(world_object.x,world_object.y,world_object.xVel,world_object.yVel, ExplosionB)
			table.remove(self.worldObjects, i) --skip all other updates for this object
			if world_object.energyIndex > -1 then self.energyObjects[world_object.energyIndex] = nil end -- it may not be in there, but whatever

		elseif world_object.status == 0 then --Remove
			if world_object.energyIndex > -1 then self.energyObjects[world_object.energyIndex] = nil end -- it may not be in there, but whatever
			table.remove(self.worldObjects, i)
			
		else
			world_object:update(dt)
	
			if world_object.collides and world_object.status > 0 then self:checkCollisions(i,world_object) end
			self:checkInBounds(world_object)
		end
	end
end

function World:describe()
	local description = ""
	description = "World: " .. #self.worldObjects .. " objects:"
	for i,world_object in pairs(self.worldObjects) do
		description = description .. ", " .. world_object.name .. " (" .. i .. ")"
	end
	return description
end

function World:isColliding(circle, i)
	i = i or 1
	local bounding_circle = {}
	for j=i+1,#self.worldObjects do
		bounding_circle.x = self.worldObjects[j].x
		bounding_circle.y = self.worldObjects[j].y
		bounding_circle.r = self.worldObjects[j].boundingRadius
		if check2Circles(circle,bounding_circle) then return true end
	end
	return false 	
end

--Check a given object against the rest of the objects in the world, starting from i
--ie. don't check objects that have already been tested themselves.
--Check bounding circles first
function World:checkCollisions(i,world_object)
	local hit_list = nil
	local object_circles = nil
	--Run collision detection 
	--We dont check objects behind us in the list, they've already been checked
		--for collisions
	local bounding_circle = {}
	local bounding_circle2 = {}
	for j=i+1,#self.worldObjects do
		world_object2 = self.worldObjects[j]
		if world_object.collides and world_object2.collides and world_object ~= world_object2 and world_object2.status > 0 then
			hit_list = nil
			bounding_circle.x = world_object.x
			bounding_circle.y = world_object.y
			bounding_circle.r = world_object.boundingRadius
		
			bounding_circle2.x = world_object2.x
			bounding_circle2.y = world_object2.y
			bounding_circle2.r = world_object2.boundingRadius
			if check2Circles(bounding_circle, bounding_circle2) then

				hit_list = circlesAreColliding( world_object:getCircles(), world_object2:getCircles())				
				if #hit_list > 0 then --hit detected
					if world_object.receivesCollisions and world_object2.givesCollisions then
						world_object:receiveCollision(hit_list, world_object2)
					end
					--print (world_object.name .. " Receiving collision from " .. world_object2.name)
					if world_object2.receivesCollisions and world_object.givesCollisions then						
						--Run this again to find out which of the receiving object's circles were hit	
						hit_list = nil
						hit_list = circlesAreColliding(world_object2:getCircles(), world_object:getCircles())		
						if hit_list == nil then
							log("Error: Collision recipient reports no hit circles in collision")
						else
								world_object2:receiveCollision(hit_list, world_object)
						end	
					end
					--Run physics results after, so we can use relative velocities in damage resolutions above
					if world_object2.mass > 0 and world_object.mass > 0 then --run "physics"
						world_object.xVel, world_object.yVel, world_object2.xVel, world_object2.yVel = 
							collideObjects(world_object.xVel, world_object.yVel, world_object.mass,
										   world_object2.xVel, world_object2.yVel, world_object2.mass)
					end 
				end
			end
		end
	end 
end

function World:checkInBounds(world_object)
	if ( world_object.x * world_object.x )  + ( world_object.y * world_object.y ) > self.radius * self.radius then
		world_object:outOfBounds()
	end
end

function World:setSpawnPoint(x, y)
	local spawn_pt = {}
	spawn_pt.x = x
	spawn_pt.y = y
	self.spawnPoints[#self.spawnPoints+1] = spawn_pt
	return #self.spawnPoints
end

function World:addObject(world_object, energy_object, draw_on_top)
	local i = 0
	if draw_on_top then
		table.insert(self.worldObjects, world_object) --Insert at end (draw last)
		i = #self.worldObjects
	else
		table.insert(self.worldObjects, 1, world_object) --Insert at beginning
		i = 1
	end
	if energy_object then
		world_object.energyIndex = #self.energyObjects+1
		self.energyObjects[world_object.energyIndex] = world_object
	end
end

function World:spawnFighter(name, spawn_point)
	local spawn_x, spawn_y 
	if spawn_point then 
		spawn_x = self.spawnPoints[spawn_point].x
		spawn_y = self.spawnPoints[spawn_point].y
	else --spawn not given
		--No spawn points defined... we'll make our own
		if #self.spawnPoints == 0 then
			--Define a square that fits inside the warn radius
			--No particular reason for this.. just makes it easy to find and x and y
			--roughly within the world 
			local spawn_rect = math.sqrt(self.warnRadius*self.warnRadius/2)
			for i=1,20 do
				local spawn_point = {}
				spawn_point.x = math.random(-spawn_rect, spawn_rect*2)
				spawn_point.y = math.random(-spawn_rect, spawn_rect*2)
				self.spawnPoints[i] = spawn_point
			end	
		end
		local tries = 1
		local i = math.random(1, #self.spawnPoints)
		local spawn_circle = {}

		--start at a random spawn point, test it, if it fails, test incrementally
		--until we find one that is clear of other objects.
		while tries < #self.spawnPoints do	
			spawn_point = i	
			spawn_circle.x = self.spawnPoints[i].x
			spawn_circle.y = self.spawnPoints[i].y
			spawn_circle.r = 300
			if not self:isColliding(spawn_circle) then
				tries = #self.spawnPoints + 1 --Break loop, we're done
			else 	
				tries = tries + 1
				i = (i + 1) % #self.spawnPoints
			end
		end
		--If this happens, we've failed to find one... good luck!
		if tries < #self.spawnPoints + 1 then
			log("Failed to find available spawn point")
		end

	end
	--create the fighter object
	local fighter = Fighter:new(self,self.spawnPoints[spawn_point].x,self.spawnPoints[spawn_point].y ,name)
	--add it to the objects list
	
	self:addObject(fighter, true)
	return fighter
end

function World:spawnBullet(from_fighter, x, y, x_vel, y_vel, heading)
	--create the fighter object
	local bullet = Bullet:new(from_fighter, x, y, x_vel, y_vel, heading)
	--add it to the objects list
	self:addObject(bullet, false)
	return bullet
end
function World:spawnMissile(from_fighter, x, y, x_vel, y_vel, heading, rVel, delay)
	local missile = Missile:new(self,from_fighter, x, y, x_vel, y_vel, heading, rVel, delay)
	--add it to the objects list
	self:addObject(missile, true)

	return missile
end

function World:spawnMarker(x, y, name)
	local marker = Marker:new(x,y,name)
	self.worldObjects[#self.worldObjects+1] = marker
	return marker
end

function World:spawnAsteroid(x, y, xVel, yVel, rVel, heading, extra)
	local asteroid = Asteroid:new(x, y, xVel, yVel, rVel, heading, extra)
	--add it to the objects list
	self.worldObjects[#self.worldObjects + 1] = asteroid
end

